"""
This file runs before tests are collected in order to set up the repository for tests.
"""

from os import sep
from pathlib import Path

# Delete all test output files before beginning so we don't pass a test based on a
# previous run's output.
for path in Path(".").rglob(f"**{sep}tests{sep}output{sep}*.txt"):
    path.unlink()
for path in Path(".").rglob(f"**{sep}tests{sep}output{sep}*.json"):
    path.unlink()



"""
Included generic parser functions which can be used by client developer commands.
"""


def string_parser(args: str) -> str:
    """
    Returns the stripped argument string without further modification.
    """
    return args.strip()


def no_arguments(args: str) -> None:
    """
    Use when a command accepts no arguments.
    """
    if args.strip():
        raise ParseError("This command takes no arguments.")
    return None


def pull_segment(args: str) -> tuple[str, str]:
    """
    Splits a string once and returns the first segment as well as the remainder.
    """
    result = args.split(maxsplit=1)
    segment = result.pop(0)
    remainder = result[0] if result else ""
    return segment, remainder


class ParseError(Exception):
    """
    Throw when there is a parsing error.
    """



"""Station class. Defines the station parameters."""

import asyncio
import logging
from asyncio import Future
from collections import defaultdict
from typing import Any, cast

from pax25.applications import BaseApplication
from pax25.applications.types import S
from pax25.ax25.address import Address
from pax25.exceptions import ConfigurationError
from pax25.frame_router import FrameRouter
from pax25.interfaces import INTERFACE_TYPES, Interface
from pax25.services.connection.service import ConnectionService
from pax25.services.digipeater import Digipeater
from pax25.types import StationConfig
from pax25.utils import default_parameters

logger = logging.getLogger(__name__)


class Station:
    """
    The main station class for Pax25. This class is intended to manage a physical
    station's digital affairs.
    """

    def __init__(self, *, config: StationConfig):
        self.name = config["name"].upper()
        self.interfaces: dict[str, Interface[Any]] = {}
        for key, value in config["interfaces"].items():
            if key.split() != [key]:
                raise ConfigurationError(
                    f"Interface keys may not have spaces. Got: {repr(key)}",
                )
            interface_cls = INTERFACE_TYPES[value["type"]]
            interface_cls = cast(type[Interface[Any]], interface_cls)
            self.interfaces[key] = interface_cls(
                name=key, settings=value["settings"], station=self
            )
        self.frame_router = FrameRouter(station=self)
        self.digipeater = Digipeater(station=self)
        self.connections = ConnectionService(station=self)
        self.application_map: defaultdict[str, dict[Address, BaseApplication[Any]]] = (
            defaultdict(lambda: {})
        )
        self.closing = False
        self.parameters = config.get("parameters", default_parameters())
        # Timers may get cancelled and the station may shut down before they would
        # be cleaned up normally. We track these tasks so if we shut down while any
        # are still active, we can reap them manually.
        self._active_future: Future[None] | None = None

    @property
    def running(self) -> bool:
        """
        Returns if the station is currently running.
        """
        if self._active_future is None:
            return False
        return not self._active_future.done()

    async def finished(self) -> None:
        """
        Await this to wait until the server is closed down programmatically.
        """
        if self._active_future is None:
            raise RuntimeError(
                "Station has not run. We cannot wait for it to finish if "
                "it never began!"
            )
        await self._active_future

    def run(self) -> None:
        """
        Runs the station. Right now this only brings up all interfaces, but it could
        handle other tasks if we need them later.
        """
        self.bring_up_interfaces()
        self._active_future = asyncio.Future()

    def bring_up_interfaces(self) -> None:
        """
        Attempts to bring up all interfaces and queue them into the event loop.
        """
        for interface in self.interfaces.values():
            interface.start()

    def get_nth_gateway(self, index: int) -> "Interface":
        """
        Get the nth gateway or raise an IndexError if not present.
        """
        if index <= 0:
            raise IndexError("Port numbers start at 1.")
        current = 0
        for interface in self.interfaces.values():
            if interface.gateway:
                current += 1
                if current == index:
                    return interface
        if not current:
            raise IndexError("No gateways available.")
        raise IndexError(f"Port {index} not found.")

    def register_app(
        self,
        app: type[BaseApplication[S]],
        *,
        interfaces: list[str],
        application_name: str | None = None,
        station_name: str | None = None,
        ssid: int | None = None,
        settings: S,
    ) -> BaseApplication[S]:
        """
        Registers a new application to the specified interfaces.
        """
        station_name = (station_name if station_name else self.name).upper()
        application_name = (
            application_name if application_name is not None else app.__class__.__name__
        )
        application_instance = app(
            name=application_name,
            settings=settings,
            station=self,
        )
        for interface in interfaces:
            if interface not in self.interfaces:
                raise ConfigurationError(
                    "Attempted to register application to non-existent "
                    f"interface, {repr(interface)}.",
                )
            new_address = Address(
                name=station_name,
                ssid=self.frame_router.next_ssid(
                    interface=self.interfaces[interface], name=station_name, ssid=ssid
                ),
            )
            self.application_map[interface][new_address] = application_instance
            self.frame_router.register_station(
                interface=self.interfaces[interface], address=new_address
            )
        return application_instance

    async def shutdown(self, stop_loop: bool = False) -> None:
        """
        Shut down all interfaces, cutting off the station.
        """
        if not self.running:
            # Already shut down.
            return
        self.closing = True
        await self.connections.shutdown()
        for interface in self.interfaces.values():
            await interface.shutdown()
        if self._active_future:
            self._active_future.set_result(None)
        self.closing = False
        if stop_loop:  # pragma: no cover
            asyncio.get_event_loop().stop()



"""
Types for the command line app.
"""

from dataclasses import dataclass
from typing import TYPE_CHECKING, TypedDict

from pax25.services.connection.connection import Connection

if TYPE_CHECKING:  # pragma: no cover
    from pax25.contrib.command.command import CommandLine


class CommandSettings(TypedDict, total=False):
    """
    Settings for the command line application.
    """

    # How many frames to keep in the framelog to relay to an administrator who is
    # monitoring.
    max_frame_log_size: int | None
    # Whether to automatically drop a user after they've disconnected from a remote.
    # Used mostly for testing, but may also work when acting as a node in the future.
    auto_quit: bool


@dataclass()
class CommandLineState:
    """
    Keeps track of the current state for connected sessions.
    """

    # Frames to be stringified and spat out at the connected user
    # when they are no longer remotely connected.
    connection: Connection | None = None


@dataclass()
class ShimSettings:
    """
    Settings for the shim application.
    """

    proxy_connection: Connection
    upstream_application: "CommandLine"



"""
Command line interface for controlling Pax25.
"""

import itertools
from collections.abc import Callable
from dataclasses import dataclass

from pax25 import Application, Station
from pax25 import __version__ as pax25_version
from pax25.applications.help import Help
from pax25.applications.parsers import ParseError, no_arguments, pull_segment
from pax25.applications.router import CommandContext, CommandRouter, CommandSpec
from pax25.applications.utils import build_columns, send_message
from pax25.ax25.address import Address
from pax25.ax25.constants import AX25_REPEATER_MAX
from pax25.ax25.frame import Frame
from pax25.ax25.matchers import MatchCall
from pax25.ax25.utils import roll_back_ssid
from pax25.contrib.command.shim import ShimApplication
from pax25.contrib.command.types import CommandLineState, CommandSettings, ShimSettings
from pax25.interfaces import FileInterface, Interface
from pax25.services.connection.connection import Connection, connection_key


def default_command_line_settings() -> CommandSettings:
    """
    Generate default settings for the command line application.
    """
    return {"max_frame_log_size": 256, "auto_quit": False}


def digipeaters_from_string(string: str) -> tuple[Address, ...]:
    """
    Given a string like 'via SOME-1,THING-2,WAT-3', determine digipeaters.
    """
    string = string.strip()
    if not string:
        return tuple()
    try:
        via, *args = string.split(maxsplit=1)
        if not via.lower() == "via" or not args:
            raise ValueError
    except ValueError as err:
        raise ValueError(
            "Digipeater list broken. Try something like 'via NAME-1,NAME-2'."
        ) from err
    string = args[0]
    addresses = tuple(
        Address.from_string(raw_address.strip()) for raw_address in string.split(",")
    )
    if len(addresses) > AX25_REPEATER_MAX:
        raise ValueError(f"Too many digipeaters. Max is {AX25_REPEATER_MAX}.")
    return addresses


def get_indexed_gateway(station: Station, index: int) -> Interface:
    """
    Wrapper for get_nth_gateway that re-raises an IndexError as a parsing error.
    """
    try:
        return station.get_nth_gateway(index)
    except IndexError as err:
        raise ParseError(str(err)) from err


@dataclass(kw_only=True, frozen=True)
class MakeConnectionSpec:
    """
    Used by the connection parser to pass around distilled arguments.
    """

    destination: Address
    digipeaters: tuple[Address, ...]
    interface: Interface


def create_connection_parser(station: Station) -> Callable[[str], MakeConnectionSpec]:
    """
    Generate a parser for the connection commands. Has to be generated since the
    available interfaces/ports are station-specific.
    """

    def connection_parser(raw_args: str) -> MakeConnectionSpec:
        raw_args = raw_args.strip()
        if not raw_args:
            raise ParseError("Need station address.")

        first_arg, remainder = pull_segment(raw_args)
        # First arg can be either a port number/alias, or the destination station.
        # If no port/alias is specified, then it grabs the default gateway (the first
        # registered interface where gateway is set to True)
        try:
            port_number = int(first_arg)
            interface = get_indexed_gateway(station, port_number)
            raw_destination, remainder = pull_segment(remainder)
        except ValueError:
            if first_arg in station.interfaces:
                interface = station.interfaces[first_arg]
                raw_destination, remainder = pull_segment(remainder)
            else:
                interface = get_indexed_gateway(station, 1)
                raw_destination = first_arg
        try:
            destination = Address.from_string(raw_destination)
        except ValueError as err:
            raise ParseError(str(err)) from err
        if remainder:
            try:
                digipeaters = digipeaters_from_string(remainder)
            except ValueError as err:
                raise ParseError(str(err)) from err
        else:
            digipeaters = tuple()
        return MakeConnectionSpec(
            destination=destination,
            digipeaters=digipeaters,
            interface=interface,
        )

    return connection_parser


class CommandLine(Application[CommandSettings]):
    """
    Command line application. Used to handle things like connecting to other nodes
    or changing settings at runtime.
    """

    connections: dict[Connection, CommandLineState]
    frame_log: list[Frame]
    auto_quit: bool
    router: CommandRouter

    def setup(self) -> None:
        """
        Initial setup.
        """
        self.connections = {}
        self.frame_log = []
        settings = default_command_line_settings()
        settings.update(self.settings or {})
        self.settings = settings
        self.auto_quit = settings["auto_quit"]
        # Need to match all frames when monitoring.
        match_call = MatchCall(matcher=lambda x, y: True, notify=self.log_frame)
        self.station.frame_router.register_matcher("monitor", match_call)
        self.router = CommandRouter()
        self.router.add(Help(self.router).spec)

        self.router.add(
            CommandSpec(
                command="connect",
                help="\r".join(
                    [
                        "Use connect to connect to an outbound station. Examples:",
                        "c FOXBOX            # Connect to a station named FOXBOX, by "
                        "default on SSID 0",
                        "c KW6FOX-3          # Connect to SSID 3 on a station named "
                        "KW6FOX",
                        "c KW6FOX-2 via BOOP # Connect to KW6FOX-2 through an "
                        "intermediary station named BOOP",
                        "c 2 KW6FOX-2 via BOOP # Use port 2 to connect to KW6FOX "
                        "through an intermediary station named BOOP",
                        "",
                        "If no port is specified, uses port 1.See also: PORTS",
                    ]
                ),
                parser=create_connection_parser(self.station),
                function=self.connect,
            ),
            CommandSpec(
                command="ports",
                help="\r".join(
                    [
                        "Usage: PORTS",
                        "",
                        "Lists all available interfaces/ports for connecting to "
                        "outbound stations, including aliases. Aliases are case "
                        "sensative.",
                    ]
                ),
                parser=no_arguments,
                function=self.ports,
            ),
            CommandSpec(
                command="quit",
                help="Closes the session.",
                aliases=("bye",),
                parser=no_arguments,
                function=self.quit,
            ),
        )

    def log_frame(self, frame: Frame, interface: Interface) -> None:
        """
        Log a frame heard by the server. If it's related to one of our connections, we
        drop it.
        """
        route_key = connection_key(
            frame.route.src.address, frame.route.dest.address, interface
        )
        if route_key in self.station.connections.table:
            # We don't log frames we're active party to.
            return
        for key, value in self.connections.items():
            if key.sudo and isinstance(key.interface, FileInterface):  # noqa: SIM102
                if not (
                    value.connection or self.connection_state_table[key]["command"]
                ):
                    # Send any outstanding.
                    self.dump_logs_to_connection(key)
                    send_message(key, str(frame))
                    return
        self.frame_log.append(frame)
        max_log_size = self.settings["max_frame_log_size"]
        if max_log_size and max_log_size < len(self.frame_log):
            self.frame_log.pop(0)

    def dump_logs_to_connection(self, connection: "Connection") -> None:
        """
        Send every log to the connection specified. Used when we disconnect and want
        to see what happened in the interim.
        """
        if not connection.sudo and isinstance(connection.interface, FileInterface):
            return
        while self.frame_log:
            frame = self.frame_log.pop(0)
            send_message(connection, str(frame))

    def on_proxy_bytes(self, connection: "Connection", message: bytes) -> None:
        """
        Called by the shim application to indicate that we need to send bytes downstream
        to the connecting station.

        We will eventually need to make this smart enough to stash bytes yet to be sent
        in the case of backgrounding the connection.
        """
        connection.send_bytes(message)

    def ports(self, connection: "Connection", _context: CommandContext[None]) -> None:
        """
        List all ports (interfaces) available for connecting outward.
        """
        headers = ["PORT", "ALIAS", "TYPE"]
        gateways: list[tuple[str, str, str]] = []
        number = 0
        for key, value in self.station.interfaces.items():
            if value.gateway:
                number += 1
                gateways.append((str(number), key, value.type))
        if not number:
            send_message(
                connection,
                "No gateways configured. You may not connect to outbound stations.",
            )
            return
        columns = build_columns(itertools.chain(headers, *gateways), num_columns=3)
        send_message(connection, "\r".join(columns))

    def on_proxy_killed(
        self, connection: "Connection", jump_connection: "Connection"
    ) -> None:
        """
        Called by the shim application to indicate that the connection has been killed.
        """
        self.connections[connection].connection = None
        send_message(
            connection, f"*** Disconnected from {jump_connection.second_party}"
        )
        self.dump_logs_to_connection(connection)
        if self.auto_quit:
            connection.disconnect()

    def set_up_connection(
        self,
        *,
        source_connection: "Connection",
        destination: Address,
        interface: Interface,
        digipeaters: tuple[Address, ...],
    ) -> None:
        """
        Set up an outbound connection to route the user through.
        """
        if source_connection.first_party == source_connection.second_party:
            # Internal connection. No need to cycle the SSID.
            source = source_connection.first_party
        else:
            source = roll_back_ssid(source_connection.first_party)
        key = connection_key(source, destination, interface)
        if key in self.station.connections.table:
            send_message(source_connection, "Route busy. Try again later.")
            return
        send_message(source_connection, f"Connecting to {destination}...")
        connection = self.station.connections.add_connection(
            # We are always the second party, and so we're the ones connecting outbound.
            first_party=source,
            second_party=destination,
            digipeaters=digipeaters,
            interface=interface,
            inbound=False,
            application=ShimApplication(
                name=f"(Shim for {key})",
                station=self.station,
                settings=ShimSettings(
                    proxy_connection=source_connection, upstream_application=self
                ),
            ),
        )
        self.connections[source_connection].connection = connection
        connection.negotiate()

    def connect(
        self,
        connection: "Connection",
        context: CommandContext[MakeConnectionSpec],
    ) -> None:
        """
        Connect to a remote station.
        """
        self.set_up_connection(
            source_connection=connection,
            interface=context.args.interface,
            destination=context.args.destination,
            digipeaters=context.args.digipeaters,
        )

    def quit(self, connection: "Connection", context: CommandContext[None]) -> None:
        """
        Closes the application.
        """
        send_message(connection, "Goodbye!")
        connection.disconnect()

    def run_home_command(self, connection: "Connection", message: str) -> None:
        """
        Match a message for a home command.
        """
        if not message:
            send_message(connection, "cmd:", append_newline=False)
            return
        self.router.route(connection, message)

    def on_startup(self, connection: "Connection") -> None:
        """
        Set up the current connection's state.
        """
        self.connections[connection] = CommandLineState()
        self.dump_logs_to_connection(connection)
        send_message(
            connection,
            f"PAX25 v{'.'.join(pax25_version)} CLI, AGPLv3\rcmd:",
            append_newline=False,
        )

    def on_message(self, connection: "Connection", message: str) -> None:
        """
        Handle command from the user.
        """
        if jump_connection := self.connections[connection].connection:
            jump_connection.send_bytes((message + "\r").encode("utf-8"))
            return
        # Send any outstanding frames that entered while they were typing before
        # giving them a response, so we're not missing any.
        self.dump_logs_to_connection(connection)
        self.run_home_command(connection, message)

    def on_shutdown(self, connection: "Connection") -> None:
        """
        Clean up connection state.
        """
        del self.connections[connection]



"""
This module contains the different interfaces which are included with pax25.
"""

from .dummy import DummyInterface
from .file import FileInterface
from .serial import SerialInterface
from .types import Interface
from .udp import UDPInterface

INTERFACE_TYPES = {
    "file": FileInterface,
    "dummy": DummyInterface,
    "serial": SerialInterface,
    "udp": UDPInterface,
}

__all__ = [
    "FileInterface",
    "Interface",
    "SerialInterface",
    "DummyInterface",
    "UDPInterface",
    "INTERFACE_TYPES",
]



"""
Serial KISS reader.
"""

import asyncio
import logging
import os
from asyncio import Queue, Task
from typing import TYPE_CHECKING

from serial import Serial

from pax25.ax25.frame import Frame
from pax25.interfaces.kiss.constants import (
    KISS_CMD_DATA,
    KISS_CMD_FULLDUP,
    KISS_CMD_PERSIST,
    KISS_CMD_SLOTTIME,
    KISS_CMD_TXDELAY,
    KISS_ENDIAN,
)
from pax25.interfaces.kiss.protocol import ax25_frames_from_kiss, kiss_command
from pax25.interfaces.types import (
    Interface,
    SerialSettings,
    TerminalNodeControllerSettings,
)
from pax25.utils import async_wrap, cancel_all

if TYPE_CHECKING:  # pragma: no cover
    from pax25.station import Station


logger = logging.getLogger(__name__)

aprint = async_wrap(print)


def default_tnc_settings() -> TerminalNodeControllerSettings:
    """
    Default settings to send to the TNC.
    """
    return {
        "port": 0,
        "tx_delay": 12,
        "persist": 63,
        "slot_time": 1,
        "full_duplex": 0,
    }


class SerialInterface(Interface[SerialSettings]):
    """
    Interface for KISS serial interaction.
    """

    type = "Serial"

    def __init__(self, name: str, settings: SerialSettings, station: "Station") -> None:
        """
        Initialize a serial interface.
        """
        self.name = name
        self.station = station
        default_path = "COM1" if os.name == "nt" else "/dev/ttyUSB0"
        self.path = settings.get("path", default_path)
        self.baud_rate = settings.get("baud_rate", 9600)
        self.timeout = settings.get("timeout", 1)
        self.write_timeout = settings.get("write_timeout", 2)
        self.rtscts = settings.get("rtscts", False)
        self.dsrdtr = settings.get("dsrdtr", False)
        self.xonxoff = settings.get("xonxoff", False)
        self.exit_kiss = settings.get("exit_kiss", False)
        self.gateway = settings.get("gateway", True)
        self.send_queue: Queue[Frame] = Queue()
        tnc_settings = settings.get("tnc_settings", {})
        tnc_settings.update(default_tnc_settings())
        self.tx_delay = tnc_settings["tx_delay"].to_bytes(1, KISS_ENDIAN)
        self.persist = tnc_settings["persist"].to_bytes(1, KISS_ENDIAN)
        self.slot_time = tnc_settings["slot_time"].to_bytes(1, KISS_ENDIAN)
        self.full_duplex = tnc_settings["full_duplex"].to_bytes(1, KISS_ENDIAN)
        self.port = tnc_settings["port"]
        self.serial: Serial = Serial(
            self.path,
            baudrate=self.baud_rate,
            bytesize=8,
            parity="N",
            stopbits=1,
            rtscts=self.rtscts,
            dsrdtr=self.dsrdtr,
            xonxoff=self.xonxoff,
        )
        self._read_loop: Task[None] | None = None
        self._write_loop: Task[None] | None = None

    @property
    def listening(self) -> bool:
        """
        Returns a bool indicating whether the interface is listening.
        """
        if not self._read_loop:
            return False
        return not self._read_loop.done()

    def start(self) -> None:
        """
        Starts the serial read loop.
        """
        if self.exit_kiss:
            self.exit_kiss_mode()
            return
        self._read_loop = asyncio.ensure_future(self.read_loop())
        self._write_loop = asyncio.ensure_future(self.write_loop())

    def send_settings(self) -> None:
        """
        Send the bytes down serial to configure the TNC.
        """
        # Need to write out the bytes to set up the TNC.
        self.serial.write(
            kiss_command(KISS_CMD_TXDELAY, self.port, self.tx_delay)
            + kiss_command(KISS_CMD_PERSIST, self.port, self.persist)
            + kiss_command(KISS_CMD_SLOTTIME, self.port, self.slot_time)
            + kiss_command(KISS_CMD_FULLDUP, self.port, self.full_duplex)
        )

    def exit_kiss_mode(self) -> None:
        """
        Can be used if we need to exit KISS mode. Note: This might
        differ for different implementations, but works for Kantronics systems.
        """
        self.serial.write(b"\xc0\xff\xc0")

    async def read_loop(self) -> None:
        """
        Connect to the given serial object and then start reading frames.
        """
        self.send_settings()
        async for frame in ax25_frames_from_kiss(
            lambda: async_wrap(self.serial.read)(1)
        ):
            self.station.frame_router.process_frame(self, frame)

    async def write_loop(self) -> None:
        """
        Write loop for serial interface.
        """
        while frame := await self.send_queue.get():
            self.serial.write(kiss_command(KISS_CMD_DATA, self.port, frame.assemble()))

    def send_frame(self, frame: Frame) -> None:
        """
        Send frames out on KISS to the serial device.
        """
        self.send_queue.put_nowait(frame)

    async def shutdown(self) -> None:
        """
        Close the serial connection and stop listening.
        """
        self._read_loop, self._write_loop = await cancel_all(
            [self._read_loop, self._write_loop]
        )
        if self.serial.is_open:
            self.serial.flush()
            self.serial.close()



"""
Quick and common client imports for Pax25.
"""

import tomllib
from pathlib import Path

from . import interfaces
from .applications.application import Application
from .station import Station

with open(Path(__file__).parent.parent / "pyproject.toml", "rb") as package_file:
    __version__ = tuple(tomllib.load(package_file)["project"]["version"].split("."))

__all__ = ["Station", "Application", "interfaces"]



"""
Types used for type checking the library. Client developers can use these if they like,
but they're mostly for our own sanity checks during development/testing.
"""

from typing import TYPE_CHECKING, NotRequired, TypedDict, Union

if TYPE_CHECKING:  # pragma: no cover
    from typing import Any

    from pax25.applications import BaseApplication
    from pax25.interfaces.types import (
        DummyInterfaceConfig,
        FileInterfaceConfig,
        SerialInterfaceConfig,
        UDPInterfaceConfig,
    )


class EmptyDict(TypedDict):
    """
    Empty dictionary.
    """


class ParameterSettings(TypedDict):
    """
    Tunable parameters. You can probably leave these alone.
    """

    retries: int
    retry_interval: int


# The StationConfig object should always be JSON-serializable. This way we'll eventually
# be able to use JSON for configuration files.


# This means that all interface configs will ALSO need to be JSON-serializable.
class StationConfig(TypedDict):
    """
    Configuration for a station.
    """

    name: str
    interfaces: dict[
        str,
        Union[
            "FileInterfaceConfig",
            "DummyInterfaceConfig",
            "SerialInterfaceConfig",
            "UDPInterfaceConfig",
        ],
    ]
    parameters: NotRequired[ParameterSettings]


# Map for the station's application registry
ApplicationMap = dict[str, dict[str, "BaseApplication[Any]"]]



"""
Connection service. Keeps track of the connection table, instantiates connections,
performs other housekeeping around them.
"""

from typing import TYPE_CHECKING, Any

from pax25.applications import BaseApplication
from pax25.ax25.address import Address
from pax25.ax25.constants import UCommand
from pax25.ax25.frame import Frame
from pax25.ax25.matchers import U_FRAME_CHECKS, MatchCall, check_all, repeats_completed
from pax25.interfaces import Interface
from pax25.services.connection.connection import (
    Connection,
    ConnectionKey,
    connection_key,
)

if TYPE_CHECKING:  # pragma: no cover
    from pax25.station import Station


class ConnectionService:
    """
    Connection handling service.
    """

    def __init__(self, station: "Station", auto_register: bool = True) -> None:
        """
        Initialize the connection service.
        """
        self.station = station
        if auto_register:
            self.register()
        self.table: dict[ConnectionKey, Connection] = {}

    def connection_matcher(self, frame: Frame, interface: Interface) -> bool:
        """
        Matches incoming connection frames.
        """
        if not check_all(
            repeats_completed, U_FRAME_CHECKS[UCommand.SET_ASYNC_BALANCED_MODE]
        )(frame, interface):
            return False
        return frame.route.dest.address in set(
            self.station.application_map.get(interface.name, [])
        )

    def register(self) -> None:
        """
        Register the matchers for the connection service.
        """

        self.station.frame_router.register_matcher(
            "connection",
            MatchCall(matcher=self.connection_matcher, notify=self.inbound_connect),
        )

    def inbound_connect(self, frame: Frame, interface: Interface) -> None:
        """
        Add a connection to the connection table based on an incoming frame.
        """
        first_party = frame.route.src.address
        second_party = frame.route.dest.address
        digipeaters = tuple(
            digipeater.address for digipeater in frame.route.digipeaters
        )
        key = connection_key(first_party, second_party, interface)
        if key in self.table:
            # Existing connection will handle this.
            return
        connection = self.add_connection(
            first_party=frame.route.src.address,
            second_party=frame.route.dest.address,
            interface=interface,
            digipeaters=digipeaters,
            inbound=True,
        )
        connection.negotiate()

    def add_connection(
        self,
        *,
        first_party: Address,
        second_party: Address,
        digipeaters: tuple[Address, ...] = tuple(),
        interface: Interface,
        inbound: bool,
        application: BaseApplication[Any] | None = None,
    ) -> Connection:
        """
        Add a connection to the connection table. Returns a connection if one already
        exists. Note that an existing connection may be outbound rather than inbound,
        or vice versa.
        """
        key = connection_key(first_party, second_party, interface)
        if key in self.table:
            raise ConnectionError(f"Connection already exists! {self.table[key]}")
        if not application and inbound:
            application = self.station.application_map[interface.name][second_party]
        connection = Connection(
            first_party=first_party,
            second_party=second_party,
            digipeaters=digipeaters,
            interface=interface,
            application=application,
            station=self.station,
            service=self,
            inbound=inbound,
            sudo=interface.sudo,
        )
        self.table[key] = connection
        return connection

    def remove_connection(self, connection: Connection) -> None:
        """
        Removes a connection from the connection table.
        """
        key = connection_key(
            connection.first_party, connection.second_party, connection.interface
        )
        del self.table[key]

    def unregister(self) -> None:
        """
        Removes the matchers for the connection service, preventing new connections.
        """
        self.station.frame_router.remove_matcher("connection")

    async def shutdown(self) -> None:
        """
        Shutdown the connection service, closing out all connections.
        """
        self.unregister()
        for connection in list(self.table.values()):
            connection.shutdown()



"""
UDP Interface for AX25 frames/connections.
"""

import asyncio
import logging
import socket
from asyncio import DatagramTransport, Queue, Task
from dataclasses import dataclass
from typing import TYPE_CHECKING

from pax25.ax25.frame import Frame
from pax25.interfaces.types import Interface, UDPSettings
from pax25.utils import cancel_all

if TYPE_CHECKING:  # pragma: no cover
    from pax25.station import Station


logger = logging.getLogger(__name__)

DEFAULT_PORT = 7773
DEFAULT_ADDRESS = "0.0.0.0"


@dataclass
class LoopSpec:
    """
    Dataclass for tracking a particular connection point.
    """

    write_loop: Task[None]
    send_queue: Queue[Frame]


@dataclass(kw_only=True, frozen=True)
class ResolvedConnectionSpec:
    """
    Variant of ConnectionSpec where we've resolved the values to addresses and ports.
    """

    ip_address: str
    port: int


class UDPInterface(Interface[UDPSettings]):
    """
    Interface for UDP Connections between stations. This interface allows you to
    bridge packet networks through the Internet. Note that you are responsible for
    making sure that all connected systems are controlled by a licenced operator.

    That means configuring your firewall properly to not allow datagrams from untrusted
    sources. Note that if you just open your firewall on a specific port, anyone will
    be able to send packets to you just like they can over the air. The Internet is a
    much larger place than the airwaves.

    Packets are sent to ALL specified connections, regardless of address, to emulate
    a broadcast domain.
    """

    type = "UDP"

    def __init__(self, name: str, settings: UDPSettings, station: "Station") -> None:
        self.name = name
        self.settings = settings
        self.connections: set[ResolvedConnectionSpec] = {
            ResolvedConnectionSpec(
                ip_address=value["ip_address"],
                port=value.get("port", DEFAULT_PORT),
            )
            for value in self.settings["connections"]
        }
        self.station = station
        self.transport: DatagramTransport | None = None
        self.read_queue: Queue[Frame] = Queue()
        self.send_queue: Queue[Frame] = Queue()
        self.gateway = settings.get("gateway", False)
        self._loops: list[LoopSpec] = []
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

    @property
    def listening(self) -> bool:
        """
        Returns if the interface is currently listening.
        """
        if self.transport:
            return not self.transport.is_closing()
        return False

    async def build_read_transport(self) -> None:
        """
        Build the read transport.
        """
        loop = asyncio.get_event_loop()
        self.transport, _ = await loop.create_datagram_endpoint(
            lambda: _Ax25UDPProtocol(
                station=self.station,
                interface=self,
            ),
            local_addr=(
                self.settings.get("listening_address", DEFAULT_ADDRESS),
                self.settings.get("listening_port", DEFAULT_PORT),
            ),
        )

    async def write_loop_for(
        self, entry: ResolvedConnectionSpec, queue: Queue[Frame]
    ) -> None:
        """
        Given a queue of Frames, loop over sending the frames.
        """
        while frame := await queue.get():
            try:
                self.sock.sendto(frame.assemble(), (entry.ip_address, entry.port))
            except Exception as err:
                logger.warning(
                    "Error sending frame to %s: %s",
                    ((entry.ip_address, entry.port), err),
                )

    def build_write_loops(self) -> None:
        """
        Build all the write loops for the specified connections.
        """
        for entry in self.connections:
            queue: Queue[Frame] = Queue()
            loop = asyncio.ensure_future(self.write_loop_for(entry, queue))
            self._loops.append(LoopSpec(send_queue=queue, write_loop=loop))

    def start(self) -> None:
        """
        Start the loops.
        """
        asyncio.ensure_future(self.build_read_transport())
        self.build_write_loops()

    def send_frame(self, frame: Frame) -> None:
        """
        Send frame to all write addresses.
        """
        for spec in self._loops:
            spec.send_queue.put_nowait(frame)

    async def shutdown(self) -> None:
        """
        Shut down all the listeners for UDP.
        """
        if self.transport:
            self.transport.close()
        await cancel_all(spec.write_loop for spec in self._loops)


class _Ax25UDPProtocol(asyncio.DatagramProtocol):
    """
    Datagram Protocol used by the UDP Interface to handle incoming packets.
    """

    def __init__(
        self,
        station: "Station",
        interface: UDPInterface,
    ):
        self.station = station
        self.interface = interface

    def datagram_received(self, data: bytes, addr: tuple[str, int]) -> None:
        """
        Handle the received datagram, disassembling the packet and sending it to
        the read queue.
        """
        try:
            frame = Frame.disassemble(data)
        except Exception as err:
            logger.info(
                "Dropping invalid frame from %s: %s",
                (addr, err),
            )
            return
        self.station.frame_router.process_frame(self.interface, frame)



"""
Utility functions/coroutines for the pax25 project.
"""

import asyncio
import os
from asyncio import CancelledError, Task
from collections.abc import Callable, Coroutine, Iterable
from contextlib import suppress
from functools import partial, singledispatch
from typing import (
    TYPE_CHECKING,
    Any,
    ParamSpec,
    TypeVar,
)

if TYPE_CHECKING:  # pragma: no cover
    from pax25.types import ParameterSettings

P = ParamSpec("P")
R = TypeVar("R")


def async_wrap(func: Callable[P, R]) -> Callable[P, Coroutine[None, None, R]]:
    """
    Wraps a function that requires syncronous operation so it can be awaited instead
    of blocking the thread.

    Shamelessly stolen and modified from:
    https://dev.to/0xbf/turn-sync-function-to-async-python-tips-58nn
    """

    async def run(*args: P.args, **kwargs: P.kwargs) -> R:
        loop = asyncio.get_event_loop()
        part = partial(func, *args, **kwargs)
        return await loop.run_in_executor(None, part)

    return run


class EnumReprMixin:
    """
    Mixin for enums that allows their export to remain copy-pastable for instantiation.
    """

    def __repr__(self) -> str:
        """
        Repr for an Enum that allows for copy-pastable instantiation.
        """
        return f"{self.__class__.__name__}.{self._name_}"  # type: ignore [attr-defined]


def default_parameters() -> "ParameterSettings":
    """
    Default parameters for parameter settings.
    """
    return {"retries": 10, "retry_interval": 5000}


def digipeater_factor(*, interval: int, hops: int, constant: int = 1000) -> int:
    """
    Returns a modified length of time based on the number of digipeaters used.

    Each digipeater increases the amount of time required between retries.
    """
    # For now, assuming this is a straight multiplication factor, though maybe we should
    # add some additional padding for processing time.
    constant = constant if hops else 0
    return (hops + 1) * interval + constant


async def cancel(task: Task[Any]) -> None:
    """
    Cancels a task, and then waits for it to be cleaned up.
    """
    task.cancel()
    with suppress(CancelledError):
        await task


async def cancel_all(tasks: Iterable[Task[Any] | None]) -> list[None]:
    """
    Cancel a series of tasks. Returns a None for each task as they are cancelled,
    useful for closing out optionally running tasks.
    """
    results: list[None] = []
    for task in tasks:
        if task is not None:
            await cancel(task)
        results.append(None)
    return results


@singledispatch
def normalize_line_endings(data: bytes) -> bytes:
    """
    Normalize line endings in a bytestring to be sensible on the target sytem.
    """
    return (
        data.replace(b"\r\n", b"\r")
        .replace(b"\n", b"\r")
        .replace(b"\r", os.linesep.encode("utf-8"))
    )



"""
Data structures and functions used for ingesting and exporting KISS frames.
"""

import logging
from collections.abc import AsyncIterator, Awaitable, Callable
from dataclasses import dataclass

from pax25.ax25.exceptions import DisassemblyError
from pax25.ax25.frame import Frame
from pax25.ax25.protocols import Assembler
from pax25.interfaces.kiss.constants import (
    KISS_CMD_DATA,
    KISS_ENDIAN,
    KISS_ESCAPED_FEND,
    KISS_ESCAPED_FESC,
    KISS_FEND,
    KISS_FESC,
    KISS_SHIFT_PORT,
    KISS_TFEND,
    KISS_TFESC,
)

logger = logging.getLogger(__name__)


@dataclass
class ReaderState:
    """
    Dataclass for keeping track of KISS frame ingestion.
    """

    packet: bytes = b""
    command: bytes = b""
    packet_started: bool = False
    reading: bool = False
    escaped: bool = False
    ready: bool = False


ReaderUpdater = Callable[[bytes, ReaderState], None]


def requires_read(func: ReaderUpdater) -> ReaderUpdater:
    """
    Makes a function inert if the 'reading' state isn't set
    to true.
    """

    def wrapped(byte: bytes, state: ReaderState) -> None:
        if not state.reading:
            return
        func(byte, state)

    return wrapped


def handle_f_end(_byte: bytes, state: ReaderState) -> None:
    """
    Handle an f-end marker.
    """
    if not state.packet:
        state.packet_started = True
        return
    state.reading = False
    state.ready = True
    return


@requires_read
def handle_tf_esc(byte: bytes, state: ReaderState) -> None:
    """
    Handle a tf escape.
    """
    if state.escaped:
        state.packet += KISS_FESC
        state.escaped = False
        return
    # Just a normal byte if we're not escaped.
    state.packet += byte


@requires_read
def handle_f_esc(_byte: bytes, state: ReaderState) -> None:
    """
    Handle an f escape.
    """
    if state.escaped:
        # This is an error. Continue.
        state.escaped = False
        return
    state.escaped = True


@requires_read
def handle_tf_end(byte: bytes, state: ReaderState) -> None:
    """
    Handle a tf_end.
    """
    if state.escaped:
        state.packet += KISS_FEND
        state.escaped = False
        return
    # Not in escaped mode, so this is the actual byte.
    state.packet += byte


@requires_read
def handle_new_char(byte: bytes, state: ReaderState) -> None:
    """
    Handle any other byte character that's part of the KISS frame.
    """
    state.packet += byte


COMMAND_FUNCS: dict[bytes, ReaderUpdater] = {
    KISS_FEND: handle_f_end,
    KISS_FESC: handle_f_esc,
    KISS_TFEND: handle_tf_end,
    KISS_TFESC: handle_tf_esc,
}


def kiss_command(command: int, port: int, value: bytes) -> bytes:
    """
    Format a kiss command for transmission.
    """
    cmd_byte = port
    cmd_byte <<= KISS_SHIFT_PORT
    cmd_byte |= command
    value = value.replace(KISS_FESC, KISS_ESCAPED_FESC)
    value = value.replace(KISS_FEND, KISS_ESCAPED_FEND)
    return KISS_FEND + cmd_byte.to_bytes(1, KISS_ENDIAN) + value + KISS_FEND


# If we ever need new frame command types for some reason, we can register them here.
KISS_FRAME_DISASSEMBLERS: dict[bytes, type[Assembler]] = {
    KISS_CMD_DATA.to_bytes(1, KISS_ENDIAN): Frame
}


async def ax25_frames_from_kiss(
    read_byte: Callable[[], Awaitable[bytes]],
) -> AsyncIterator[Frame]:
    """
    Generator loop for reading KISS from a source. Drops all frames except for AX.25
    frames.
    """
    async for _, frame in read_from_kiss(read_byte):
        match frame:
            case Frame():
                yield frame


async def read_from_kiss(
    read_byte: Callable[[], Awaitable[bytes]],
) -> AsyncIterator[tuple[int, Assembler]]:
    """
    Generator loop for reading KISS frames from a source.
    """
    state = ReaderState()
    while byte := await read_byte():
        if state.packet_started and not state.reading and byte != KISS_FEND:
            # This is the command byte.
            state.command = byte
            state.reading = True
            continue
        modifier: ReaderUpdater = COMMAND_FUNCS.get(byte, handle_new_char)
        modifier(byte, state)
        if state.ready:
            try:
                disassembler = KISS_FRAME_DISASSEMBLERS.get(state.command, None)
                if disassembler is None:
                    raise DisassemblyError(
                        "No existing frame disassembler for this command."
                    )
                frame = disassembler.disassemble(state.packet)
                yield int.from_bytes(state.command, KISS_ENDIAN), frame
            except Exception as err:
                logger.debug("Packet decoding error: %s", (err,))
            state = ReaderState()



"""
CommandRouter module. We use the command router to route commands for the command line
app. Other apps can use it, too, as it's not strictly bound to the command line app.
"""

from collections.abc import Callable
from dataclasses import dataclass
from typing import Any

from pax25.applications.autocomplete import AutocompleteDict
from pax25.applications.parsers import ParseError, string_parser
from pax25.applications.utils import send_message
from pax25.exceptions import ConfigurationError
from pax25.services.connection.connection import Connection

type CommandFunc[P] = Callable[[Connection, "CommandContext[P]"], None]

type CommandMap = dict[str, CommandFunc[object]]


@dataclass(frozen=True, kw_only=True)
class CommandSpec[P]:
    """
    Used for defining commands that are used by the command line app, or any other app
    that follows its conventions for commands.
    """

    command: str
    aliases: tuple[str, ...] = tuple()
    # A long description of the command and its usage.
    help: str
    # The function for this command, when run.
    function: CommandFunc[P]
    parser: Callable[[str], P]


@dataclass()
class CommandContext[P]:
    """
    Context handed to a command.
    """

    spec: CommandSpec[P]
    # The actual command as entered by the user.
    command: str
    args: P


def default_command_func(connection: Connection, context: CommandContext[str]) -> None:
    """
    Default command used by CommandRouter.
    """
    send_message(connection, f"{repr(context.command)} is not a recognized command.")


default_command = CommandSpec[str](
    command="",
    help="",
    function=default_command_func,
    parser=string_parser,
)


class CommandRouter:
    """
    Router object that allows us to quickly route to command functions based on a
    command string sent by a user.
    """

    def __init__(
        self,
        default: CommandSpec[Any] = default_command,
    ) -> None:
        """
        Creates an autocompleting command map that handles
        """
        self.command_store: AutocompleteDict[CommandSpec[Any]] = AutocompleteDict()
        # Canonical listing of all command names, used for checking conflicts.
        self.command_set: set[str] = set()
        # Canonical listing of all aliases, used for checking conflicts.
        self.alias_set: set[str] = set()
        self.default = default

    @property
    def help_available(self) -> bool:
        """
        Returns a boolean indicating if there's a help command installed.
        """
        try:
            results = self.command_store["HELP"]
            # Help command is ambiguous if there's more than 1.
            # This could still be wrong if there is an entry named something like
            # 'helpmeplz' and it's not a help command. But we're going to discount that
            # possibility here. If it ever becomes a real problem we'll refactor this
            # to something more robust.
            return len(results) == 1
        except KeyError:
            return False

    def add(self, *args: CommandSpec[Any]) -> None:
        """
        Add commands to the command router.
        """
        for arg in args:
            command = arg.command.upper()
            aliases = set(alias.upper() for alias in arg.aliases)
            to_check = (command, *aliases)
            for entry in to_check:
                if entry in self.command_set or entry in self.alias_set:
                    existing_spec = self.command_store[entry]
                    raise ConfigurationError(
                        f"Found preexisting entry with conflicting name or "
                        f"aliases when adding spec {repr(arg)}. Conflicting "
                        f"entry was: {repr(existing_spec)}"
                    )
            for entry in to_check:
                self.command_store[entry] = arg
            self.command_set |= {command}
            self.alias_set |= aliases

    def remove(self, *args: CommandSpec[Any]) -> None:
        """
        Remove commands from the command router.
        """
        for arg in args:
            command = arg.command.upper()
            if command not in self.command_store.store:
                raise KeyError(f"Command does not exist, {repr(arg.command)}")
            if self.command_store.store[command] != arg:
                raise KeyError(
                    f"Command {repr(arg.command)} exists, but is for a different spec!"
                )
            del self.command_store[command]
            self.command_set -= {command}
            for alias in arg.aliases:
                del self.command_store[alias.upper()]
                self.alias_set -= {alias.upper()}

    def route(self, connection: Connection, command: str) -> None:
        """
        Routes a user to a command function based on their selection, or gives them
        a hint otherwise.
        """
        segments = command.split(maxsplit=1)
        raw_command = segments[0]
        command = raw_command.upper()
        args = ""
        if len(segments) == 2:
            args = segments[1]
        try:
            candidates = self.command_store[command]
        except KeyError:
            context = CommandContext(command=raw_command, args=args, spec=self.default)
            self.default.function(connection, context)
            return
        if len(candidates) > 1:
            possibilities = sorted(entry[1].command for entry in candidates)
            send_message(
                connection,
                "Ambiguous command. "
                f"Did you mean one of these?: {', '.join(possibilities)}",
            )
            return
        [selection] = candidates
        (_key, spec) = selection
        try:
            parsed_args = spec.parser(args)
            context = CommandContext(command=raw_command, args=parsed_args, spec=spec)
        except ParseError as err:
            error_string = str(err)
            if self.help_available:
                error_string += f"\rTry: help {spec.command}"
            send_message(connection, error_string)
            return
        selection[1].function(connection, context)



"""
The Frame Router routes frames to relevant matching functions used to perform any
protocol work.
"""

from logging import getLogger
from typing import TYPE_CHECKING, Any

from pax25.ax25.address import Address
from pax25.ax25.frame import Frame
from pax25.ax25.matchers import MatchCall
from pax25.exceptions import ConfigurationError

if TYPE_CHECKING:  # pragma: no cover
    from pax25.interfaces import Interface
    from pax25.station import Station


log = getLogger(__name__)


class FrameRouter:
    """
    The Frame queue is a sort of router. It takes in frames from the varying interfaces
    and makes sure they go to their proper destination, whether that be an existing
    connection object or re-transcribed and sent out to another interface.
    """

    def __init__(self, *, station: "Station"):
        self.station = station
        # We keep track of the addresses we are internally watching for here, such
        # as contacts to our internally registered applications.
        self.internal_addresses: dict[Address, set[Interface[Any]]] = {}
        self.matchers: dict[str, MatchCall] = {}

    def process_frame(self, interface: "Interface", frame: Frame) -> None:
        """
        Interfaces call this function to put a new frame in the queue, to be interpreted
        as a received frame.

        This function must be resilient to avoid bringing down the station from one bug.
        """
        # Could change while iterating, since we might do something like register a
        # connection.
        matchers = list(self.matchers.items())
        for key, match_call in matchers:
            try:
                if not match_call.matcher(frame, interface):
                    continue
            except Exception:
                log.exception(
                    "Got error for matcher with key %s:",
                    repr(key),
                )
                continue
            try:
                match_call.notify(frame, interface)
            except Exception:
                log.exception(
                    "Got error for notifier with key %s:",
                    repr(key),
                )
                continue

    def send_frame(self, interface: "Interface", frame: Frame) -> None:
        """
        Send a frame out on a specific interface. For now, there's no advantage to
        using this over the interface's send_frame function, but there may later
        be filters which watch for frames that should be removed.
        """
        interface.send_frame(frame)

    def register_matcher(self, key: str, match_call: MatchCall) -> None:
        """
        Registers a matcher.
        """
        if key in self.matchers:
            log.warning(
                "Existing matcher for key %s, %s, was replaced with new matcher %s. "
                "This may be a collision or a sign of a cleanup issue.",
                repr(key),
                self.matchers[key],
                match_call,
            )
        self.matchers[key] = match_call

    def remove_matcher(self, key: str) -> None:
        """
        Removes a matcher based on its key.
        """
        if key not in self.matchers:
            log.warning(
                "Non-existent key removed, %s. This may indicate that the key was "
                "never registered, or cleanup has been called multiple times "
                "unnecessarily.",
                repr(key),
            )
            return
        del self.matchers[key]

    def next_ssid(
        self,
        *,
        name: str,
        interface: "Interface[Any]",
        ssid: int | None = None,
    ) -> int:
        """
        Given a station name and an interface, produces the next SSID.

        You can give an ssid you would prefer by specifying the ssid argument. If the
        ssid already exists or is invalid, throws an exception. This function is
        used primarily by the station in order to assign SSIDs to applications.
        """
        name = name.upper()
        ssids: set[int] = set()
        for address, interfaces in self.internal_addresses.items():
            if address.name != name:
                continue
            if interface in interfaces:
                ssids |= {address.ssid}
        if ssid is None:
            for candidate in range(16):
                if candidate not in ssids:
                    ssid = candidate
                    break
            else:
                raise ConfigurationError(
                    f"All valid SSIDs on {name}, interface {interface} are taken! "
                    "You may have at most 16 SSIDs, numbered 0-15."
                )
        if not 0 <= ssid <= 15:
            raise ConfigurationError("SSID must be between 0 and 15.")
        if ssid in ssids:
            raise ConfigurationError(
                f"SSID {repr(ssid)} already registered on {interface}!"
            )
        return ssid

    def register_station(
        self, *, interface: "Interface[Any]", address: Address
    ) -> None:
        """
        Registers an address as existing on a particular interface. It can then be
        looked up later for making connections.
        """
        self.internal_addresses[address] = self.internal_addresses.get(
            address, set()
        ) | {interface}
